#include "Server.h"
#include <iostream>
#include <thread>
#include <chrono>
#include <algorithm>
#include <cstdlib>

Server::Server(unsigned short port) {
    listenSock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(listenSock == INVALID_SOCKET) {
        std::cerr << "Failed to create socket." << std::endl;
        std::exit(EXIT_FAILURE);
    }
    sockaddr_in serverAddr;
    serverAddr.sin_family = AF_INET;
    serverAddr.sin_addr.s_addr = INADDR_ANY;
    serverAddr.sin_port = htons(port);
    if(bind(listenSock, reinterpret_cast<sockaddr*>(&serverAddr), sizeof(serverAddr)) == SOCKET_ERROR) {
        std::cerr << "Bind failed." << std::endl;
        closesocket(listenSock);
        std::exit(EXIT_FAILURE);
    }
    if(listen(listenSock, SOMAXCONN) == SOCKET_ERROR) {
        std::cerr << "Listen failed." << std::endl;
        closesocket(listenSock);
        std::exit(EXIT_FAILURE);
    }
}

Server::~Server() {
    closesocket(listenSock);
}

void Server::start() {
    std::thread([this](){
        while(true) {
            SOCKET clientSock = accept(listenSock, nullptr, nullptr);
            if(clientSock == INVALID_SOCKET) {
                std::cerr << "Accept failed." << std::endl;
                continue;
            }
            std::cout << "Client connected." << std::endl;
            {
                std::lock_guard<std::mutex> lock(mutex_);
                clients.push_back(clientSock);
            }
            // Broadcast join packet.
            Packet joinPacket;
            joinPacket.type = JOIN;
            joinPacket.data = "Client Connected";
            broadcast(joinPacket);
            // Spawn a thread to handle the new client.
            std::thread(&Server::handleClient, this, clientSock).detach();
        }
    }).detach();
}

void Server::broadcast(const Packet &packet) {
    std::lock_guard<std::mutex> lock(mutex_);
    for(auto it = clients.begin(); it != clients.end();) {
        if(!sendPacket(*it, packet)) {
            // Remove client if send fails.
            closesocket(*it);
            it = clients.erase(it);
        } else {
            ++it;
        }
    }
}

void Server::handleClient(SOCKET clientSock) {
    bool authenticated = false;
    std::string sessionToken;
    Packet packet;
    
    while(receivePacket(clientSock, packet)) {
        if(packet.type == AUTH_REQUEST) {
            std::string creds = packet.data;
            size_t delim = creds.find(":");
            Packet authResp;
            authResp.type = AUTH_RESPONSE;
            if(delim != std::string::npos) {
                std::string username = creds.substr(0, delim);
                std::string password = creds.substr(delim + 1);


                if(username == "testuser" && password == "pass123") {

                    authenticated = true;
                    sessionToken = "SESSION_TOKEN_ABC123";
                    authResp.data = sessionToken;
                    std::cout << "Client authenticated: " << username << std::endl;

                } else {
                    authResp.data = "AUTH_FAILED";
                    std::cout << "Client failed to authenticate: " << username << std::endl;
                }

 

            } else {

                authResp.data = "AUTH_FAILED";
                std::cerr << "Invalid authentication packet format." << std::endl;

            }
            sendPacket(clientSock, authResp);
        }
        else if(!authenticated) {


            Packet notAuth;
            notAuth.type = AUTH_RESPONSE;
            notAuth.data = "NOT_AUTHENTICATED";
            sendPacket(clientSock, notAuth);


        }
        else {

            switch(packet.type) {
                case MOVE:
                    broadcast(packet);
                    break;
                case HEARTBEAT:
                    break;
                case JOIN:
                    std::cout << "Join packet received: " << packet.data << std::endl;
                    break;
                case DISCONNECT:
                    std::cout << "Disconnect packet received: " << packet.data << std::endl;
                    break;
                default:
                    std::cout << "Unknown packet type: " << static_cast<int>(packet.type)
                              << " | Data: " << packet.data << std::endl;
                    break;
            }
        }
    }
    std::cout << "Client disconnected." << std::endl;
    Packet disconnectPacket;
    disconnectPacket.type = DISCONNECT;
    disconnectPacket.data = "A client disconnected.";
    broadcast(disconnectPacket);
    std::lock_guard<std::mutex> lock(mutex_);
    auto it = std::find(clients.begin(), clients.end(), clientSock);
    if(it != clients.end()) {
        closesocket(*it);
        clients.erase(it);
    }
}
